package org.edg.wl.dgas.dgassim;

import java.util.*;
import java.io.*;

import org.edg.wl.dgas.dgassim.workload.pa.*;
import org.edg.wl.dgas.dgassim.workload.rb.*;
import org.edg.wl.dgas.dgassim.monitor.*;
import org.edg.wl.dgas.dgassim.fabric.*;

public class AutoJobSubmSimCore extends Thread implements SimCore {

    private boolean verbose;
    private boolean debug;

    private static int MSEC_BETWEEN_STDOUT, MSEC_BETWEEN_JOBS;

    private static int JOBS_PER_SUBMISSION;

    private static String PRICE_OUTFILE_PREFIX, QWT_OUTFILE_PREFIX;

    private FileOutputStream jobsubmOutFile;

    public static long maxSimRunTime; // max. duration (in miliseconds)

    public long timeOfSimStart; // timestamp of start of simulation
    public long timeOfLastComp; // timestamp of last computation
    public long timeOfLastJob; // timestamp of last job sumbission

    public long totalCompEnergy; // computational energy requested by all
                                 // jobs during run time.
    public long totalCompEnSubmitted; // total comp. en. of all submitted jobs

    public GIS gis;
    public ResourceBroker rb;

    private int lastJobId, noJobsRefused;

    private boolean runSim;

    public AutoJobSubmSimCore(GIS gis, ResourceBroker rb, long msMaxRunTime,
			      String priceOutfilePrefix,
			      String qwtOutFilePrefix,
			      String jobsubout,
			      int msStdout, int msJobs, int jobsPerSubm,
			      boolean verbose, boolean debug) {
	super();

	timeOfLastComp = (new Date()).getTime();
	timeOfLastJob = (new Date()).getTime();

	this.gis = gis;
	this.rb = rb;

	this.verbose = verbose;
	this.debug = debug;

	PRICE_OUTFILE_PREFIX = priceOutfilePrefix;
	QWT_OUTFILE_PREFIX = qwtOutFilePrefix;

	try {
	    jobsubmOutFile = new FileOutputStream(jobsubout);
	    jobsubmOutFile.write(new String("# Job submission logfile\n"
					    + "# the boolean number at the "
					    + "beginning of each records "
					    + "indicates,\n"
					    + "# whether the job could be "
					    + "submitted or not\n")
		                                                 .getBytes());
	} catch (Exception e) {
	    System.err.println("Cannot create output file for job submission "
			       + "log");
	    System.exit(-1);
	}

	MSEC_BETWEEN_STDOUT = msStdout;
	MSEC_BETWEEN_JOBS = msJobs;
	JOBS_PER_SUBMISSION = jobsPerSubm;

	maxSimRunTime = msMaxRunTime;

	if (debug)
	    System.out.println("constructor of class AutoJobSubmSimCore");

	if (debug)
	    System.out.println("MSEC_BETWEEN_STDOUT="+MSEC_BETWEEN_STDOUT
			       +", MSEC_BETWEEN_JOBS="+MSEC_BETWEEN_JOBS
			       +", JOBS_PER_SUBMISSION="+JOBS_PER_SUBMISSION
			       +", maxSimRunTime="+maxSimRunTime);
    } // constructor

    public void initSimulation() {

	totalCompEnergy = 0;
	totalCompEnSubmitted = 0;
	lastJobId = 0;
	noJobsRefused = 0;

	ArrayList ceList = gis.getCEList();

	for (int i = 0; i < ceList.size(); i++) {

	    ComputingElement ce = (ComputingElement)ceList.get(i);
	    int id = ce.getId();

	    try {
		ce.setQWTOutFile(new FileOutputStream(QWT_OUTFILE_PREFIX
						      +"_CE"+id+".dat"));
		ce.getQWTOutFile().write(
	            (new String("# current QWT(time units) of CE"+id
				+ " (as function of time(s))\n"
				+ "# time QWT\n"
				+ "0 "+ce.getCurrQWTNoComp()+"\n")).getBytes()
		    );
		ce.setPriceOutFile(new FileOutputStream(PRICE_OUTFILE_PREFIX
							+"_CE"+id+".dat"));
		ce.getPriceOutFile().write(
		    (new String("# current Price(GridCredits) of CE"+id
				+ " (as function of time(s))\n"
				+ "# time Price\n"
				+ "0 "+ce.getPrice()+"\n")).getBytes()
		    );
	    } catch (Exception e) {
		System.err.println("Cannot create output files for CE"+id);
		System.exit(-1);
	    }
	}
    }

    public void runSimulation() {

	timeOfSimStart = (new Date()).getTime();

	// start AutoJobSubmSimCore
	start();
    }


    public void start() {
	runSim = true;
	super.start();

	System.out.println("Starting simulation with AutoJobSubmSimCore for "
			   + maxSimRunTime + " ms = "
			   + ((double)maxSimRunTime/60000) + " min.");
	if (verbose) System.out.println("Starting AutoJobSubmSimCore thread");

	// starting all PA threads
	ArrayList paList = gis.getPAList();
	for (int i = 0; i < paList.size(); i++)
	    ((PriceAuthority)paList.get(i)).start();
    }


    public void stopSimulation() {
	runSim = false;

	// stopping PA threads
	ArrayList paList = gis.getPAList();
	for (int i = 0; i < paList.size(); i++)
	    ((PriceAuthority)paList.get(i)).stopPA();

	long currTime = (new Date()).getTime();
	long seconds = (currTime - timeOfSimStart) / 1000;
	System.out.println("stopping at " + seconds + " sec.\n"
			   + lastJobId
			   + " job submission requests with a total comp. "
			   + "energy of " + totalCompEnergy + " UCE = "
			   + (totalCompEnergy/seconds) + " UCE/s.\n"
			   + (lastJobId-noJobsRefused)
			   + " requests satisfied for a total comp. energy of "
			   + totalCompEnSubmitted + " UCE = "
			   + (totalCompEnSubmitted/seconds) + " UCE/s.");
    }

    public void run() {

	long currTime;

	int jobID = 1;

	int price, cQWT;

	ArrayList ces = gis.getCEList();
	ArrayList ceIDs = gis.getCEIDList();

	while(runSim) {

	    // first check if we need to submit a job:
	    currTime = (new Date()).getTime();

	    // maximum run time reached?
	    if (maxSimRunTime != 0) // limit specified
		if (currTime >= maxSimRunTime + timeOfSimStart)
		    stopSimulation();

	    // submit jobs?
		if ( (currTime - timeOfLastJob) >= MSEC_BETWEEN_JOBS) {

		    boolean submitted;

		    for (int j = 0; j < JOBS_PER_SUBMISSION; j++) {

			// creating at random one of the job types
			Job newJob = gis.createNewJob();

			// trace consumed comp. energy
			totalCompEnergy += newJob.compEnergy();

			submitted = rb.submitJob(newJob);


			// write to job submission logfile:
			try {
			    if (submitted)
				jobsubmOutFile.write(new String("1 ")
				                                 .getBytes());
			    else
				jobsubmOutFile.write(new String("0 ")
				                                 .getBytes());

			    // in any case:
			    jobsubmOutFile.write(new String("timestamp="
				  + currTime + " jobId=" + newJob.dgJobId()
				  + " type=" + newJob.jobType() + " compEn="
				  + newJob.compEnergy() + " reqCPU="
				  + newJob.minCPUPower()).getBytes());

			    if (submitted) 
				jobsubmOutFile.write(new String(" CEId="
							 +newJob.executingCE())
			                                          .getBytes());

			    if (submitted)
				totalCompEnSubmitted += newJob.compEnergy();
			    else
				noJobsRefused++;

			    // terminate record
			    jobsubmOutFile.write(new String("\n").getBytes());

			}
			catch (Exception e) {
			    System.err.println("Cannot write to job submission"
					       + " logfile.");
			}

			lastJobId = newJob.dgJobId();

			if (verbose && !submitted)
			    System.out.println("Job " + newJob.dgJobId()
					       + " of type " + newJob.jobType()
					       + " NOT "
					       + "submitted: CPU power not "
					       + "sufficient or queues full.");
		    }
		    timeOfLastJob = (new Date()).getTime();
		}

	    // now monitor QWTs and prices ...

	    currTime = (new Date()).getTime();
	    if ( (currTime - timeOfLastComp) >= MSEC_BETWEEN_STDOUT ) {
		for (int i = 0; i < ceIDs.size(); i++) {
		    ComputingElement ce = (ComputingElement)
			      gis.getCE(((Integer)ceIDs.get(i)).intValue());
		    price = ce.getPA().getPrice(ce.getId());
		    cQWT = ce.getCurrQWT();
		    if (verbose) {
			System.out.print("CE " + i + ": p=" + price
					 + ", cQWT=" + cQWT + "; ");
			if (i == ceIDs.size()-1)
			System.out.print("\n");
		    }
		}
		timeOfLastComp = currTime;
	    }
	}
    }

}
