package org.edg.wl.dgas.dgassim;

import java.io.*;
import java.util.ArrayList;
import org.edg.wl.dgas.dgassim.workload.pa.*;
import org.edg.wl.dgas.dgassim.workload.rb.*;
import org.edg.wl.dgas.dgassim.monitor.*;
import org.edg.wl.dgas.dgassim.fabric.*;

public class DGASSim {

    public static final String VERSION = "0.0.2"; 

    public Configuration conf;
    private boolean verbose, debug;

    public SimCore simulator;
    public GIS gis;
    public PriceSensitiveUtilityModule utmod;
    public ResourceBroker rb;
    public PricingModule pmod;
    public PriceAuthority pa;

    // default values for PAs:
    public static final int DEF_WAIT_TIME = 10 * 1000; // wait time
                                                       // in miliseconds
    public static final int DEF_PRICE_ADJ_THRES = 0; // min. price difference

    // default values for pricing modules:
    public final static String DEF_PROFIT_OUTFILE_PREFIX = "profit";

    // default values for CEs:
    private final static double DEF_SEC_PER_TIME_UNIT = 0.5;

    // default values for AutoJobSubmSimCore:
    public final static int DEF_MSEC_BETWEEN_STDOUT = 2000;
    public final static int DEF_MSEC_BETWEEN_JOBS   = 2000;
    public final static int DEF_JOBS_PER_SUBMISSION = 2;
    public final static String DEF_PRICE_OUTFILE_PREFIX = "price";
    public final static String DEF_QWT_OUTFILE_PREFIX = "currQWT";
    public final static String DEF_JOBSUBMISSION_OUTFILE 
	= "jobsubmission.dslog";


    // Constructor
    public DGASSim(String confFName) {

	// ******************
	// configuration file
	// ******************

	if (confFName == null)
	    confFName = new String("etc/dgas-sim.conf");
	try {
	    conf = new Configuration(confFName);
	}
	catch (IOException e) {
	    System.err.println("Cannot read configuration file '" + confFName
			       + "'\nError: " +e.getMessage());
	    System.exit(-1);
	}

	// ***************
	// general options
	// ***************

	Boolean vb = conf.getBooleanValue("VERBOSE");
	if (vb == null)
	    verbose = false;
	else
	    verbose = vb.booleanValue();

	Boolean db = conf.getBooleanValue("DEBUG");
	if (db == null)
	    debug = false;
	else
	    debug = db.booleanValue();

	if (debug)
	    System.out.println("constructor of class DGASSim");

	// ********************************
	// create simulated grid components
	// ********************************

	// information system
	createGIS();

	// resource brokers
	createRBs();

	// price authorities
	createPAs();

	// queue types
	createQueueTypes();

	// computing elements
	createCEs();

	// job prototypes
	createJobPrototypes();

	// core of simulation
	createSimCore();

	if (debug)
	    System.out.println("CONFIGURATION: "+conf);

    } // construtcor DGAS-Sim() ...

    public static void helpMessage() {
	System.out.println("DGASSim v"+VERSION+"\n"
			   +"written 2003 by Rosario Piro (piro@to.infn.it)\n"
			   +"Istituto Nazionale di Fisica Nucleare - Torino, "
			   +"Italy.\n"
			   +"Distributed under the terms of the EDG DataGrid "
			   +"Software License,\nsee file 'LICENSE'!\n\n"
			   +"Usage: 'java DGASSim [conf-file]'\n"
			   +"   or: 'start_DGAS-Sim.sh [conf-file]'\n"
			   +"        default conf-file: 'etc/dgas-sim.conf'\n");
	System.exit(0);
    }


    // *******************
    // information service
    // *******************

    private void createGIS() {
	gis = new GIS(verbose, debug);
    }

    // ************************************
    // utility modules for resource brokers
    // ************************************

    private void createRBs() {

	// for now we support only one resource broker of type PS:

	String utilmodule = conf.getStringValue("UTILITY_MODULE");
	// default is PS
	utmod = new PriceSensitiveUtilityModule();

	// resource broker
	rb = new ResourceBroker(gis, utmod, verbose, debug);
    }


    // *****************
    // Price Authorities
    // *****************

    private void createPAs() {

	// general parameters for all PAs:

	int paWaitTime, priceAdjThres;
	Integer waitTime = conf.getIntegerValue("PA_WAIT_TIME");
	Integer prAdTh = conf.getIntegerValue("PRICE_ADJ_THRES");

	if (waitTime == null) {
	    paWaitTime = DEF_WAIT_TIME;
	    conf.addParameter("PA_WAIT_TIME",
			      new Integer(DEF_WAIT_TIME));
	}
	else
	    paWaitTime = waitTime.intValue();

	if (prAdTh == null) {
	    priceAdjThres = DEF_PRICE_ADJ_THRES;
	    conf.addParameter("PRICE_ADJ_THRES",
			      new Integer(DEF_PRICE_ADJ_THRES));
	}
	else
	    priceAdjThres = prAdTh.intValue();

	// single PAs:

	Integer numberPAs = conf.getIntegerValue("NUMBER_PA");
	if (numberPAs == null) {
	    System.err.println("Number of PAs not specified in conf file!");
	    System.exit(-1);
	}

	int nPAs = numberPAs.intValue();
	for (int id = 0; id < nPAs; id++) {
	    try {
		String paParam = conf.getStringValue("PA"+id);
		int index1 = 0, index2 = 0;
		index2 = paParam.indexOf(',');
		String paIdString = paParam.substring(index1, index2);
		index1 = index2+1;
		index2 = paParam.indexOf(',', index1);
		String pModule = paParam.substring(index1, index2);
		index1 = index2+1;
		index2 = paParam.indexOf(',', index1);

		// first create pricing module for PA:

		if (pModule.equals("HYBRID")) {
		    // *** Hybrid PM:
		    int basePrice = Integer.parseInt(paParam.substring(index1,
								      index2));
		    int deltaP = Integer.parseInt(paParam.substring(index2+1,
							    paParam.length()));

		    if (verbose)
			System.out.println("creating PA '"+paIdString
					   +"' with "+pModule
					   +" module: base price is "+basePrice
					   +", variation limit is "+deltaP);

		    pmod = new HybridPricingModule(basePrice, deltaP, verbose,
						   debug);

		}
		else if (pModule.equals("DF")) {

		    // *** Derivative Follower PM:
		    int minPrice = Integer.parseInt(paParam.substring(index1,
								      index2));
		    index1 = index2+1;
		    index2 = paParam.indexOf(',', index1);
		    double minStep = Double.parseDouble(paParam
							.substring(index1,
								      index2));
		    index1 = index2+1;
		    index2 = paParam.indexOf(',', index1);
		    double maxStep = Double.parseDouble(paParam
							.substring(index1,
								     index2));
		    index1 = index2+1;
		    index2 = paParam.indexOf(',', index1);
		    double initStep = Double.parseDouble(paParam
							.substring(index1,
								     index2));
		    index1 = index2+1;
		    index2 = paParam.indexOf(',', index1);		
		    double stepIncr = Double.parseDouble(paParam
							 .substring(index1,
								      index2));
		    index1 = index2+1;
		    index2 = paParam.indexOf(',', index1);		
		    double stepDecr = Double.parseDouble(paParam
							 .substring(index1,
								      index2));
		    double stepIncr2T = Double.parseDouble(paParam
						     .substring(index2+1,
							   paParam.length()));

		    if (verbose)
			System.out.println("creating PA '"+paIdString
					   +"' with "+pModule
					   +" module: min.price="+minPrice
					   +", min.step="+minStep
					   +", max.step="+maxStep
					   +", init.step="+initStep
					   +", step incr="+stepIncr
					   +", step decr="+stepDecr
					   +", step incr2T="+stepIncr2T);

		    String ProfitOutfilePrefix 
			= conf.getStringValue("PROFIT_OUTFILE_PREFIX");
		    if (ProfitOutfilePrefix == null)
			ProfitOutfilePrefix
			    = new String(DEF_PROFIT_OUTFILE_PREFIX);

		    pmod = new DFPricingModule(minPrice, minStep, maxStep,
					       initStep, stepIncr, stepDecr,
					       stepIncr2T, ProfitOutfilePrefix,
					       verbose, debug);
		}
		else {
		    // unknown pricing module:
		    System.err.println("Error: Unknown pricing module "
				       +pModule+" in configuration file!");
		    System.exit(-1);
		}

		// now create PA:

		pa = new PriceAuthority(id, paIdString, pmod, paWaitTime,
					priceAdjThres, verbose, debug);
		// register PA to GIS:
		gis.registerPA(pa);
	    } 
	    catch (Exception e) {
		System.err.println("PA"+id+" is not correctly defined in the "
				   + "configuration file!");
		System.exit(-1);
	    }
	}

    }


    // ***********
    // queue Types
    // ***********

    private void createQueueTypes() {

	Integer numberQTypes = conf.getIntegerValue("NUMBER_QTYPE");
	if (numberQTypes == null) {
	    System.err.println("Number of queue types not specified in conf"
			       +" file!");
	    System.exit(-1);
	}

	int nQTs = numberQTypes.intValue();
	for (int id = 0; id < nQTs; id++) {
	    try {
		String qtParam = conf.getStringValue("QTYPE"+id);
		int index1 = 0, index2 = 0;
		index2 = qtParam.indexOf(',');
		String qtIdString = qtParam.substring(index1, index2);
		int maxQWT = Integer.parseInt(qtParam.substring(index2+1,
							   qtParam.length()));
		if (verbose)
		    System.out.println("creating queue type '"+qtIdString
				       +"' with maxQWT="+maxQWT);

		gis.registerQType(qtIdString, maxQWT);
	    }
	    catch (Exception e) {
		System.err.println("QTYPE"+id+" is not correctly defined in "
				   + "the configuration file!");
		System.exit(-1);

	    }
	}
    }


    // *****************
    // Computing Element
    // *****************

    private void createCEs() {

	double secPerTU;
	Double secTU = conf.getDoubleValue("SEC_PER_TIME_UNIT");

	if (secTU == null) {
	    secPerTU = DEF_SEC_PER_TIME_UNIT;
	    conf.addParameter("SEC_PER_TIME_UNIT",
			      new Double(DEF_SEC_PER_TIME_UNIT));
	}
	else
	    secPerTU = secTU.doubleValue();

	// pa, id, cpuP (in UCE/time unit), mQWT, cQWT (both in time units)
	// computing elements become automatically registered to the PA,
	// but still have to be registered to the GIS.

	Integer numberCEs = conf.getIntegerValue("NUMBER_CE");
	if (numberCEs == null) {
	    System.err.println("Number of CEs not specified in conf file!");
	    System.exit(-1);
	}

	int nCEs = numberCEs.intValue();
	for (int id = 0; id < nCEs; id++) {
	    try {
		String ceParam = conf.getStringValue("CE"+id);

		int index1 = 0, index2 = 0;
		index2 = ceParam.indexOf(',');
		int cpuP = Integer.parseInt(ceParam.substring(index1, index2));

		index1 = index2+1;
		index2 = ceParam.indexOf(',', index1);		
		int cQWT = Integer.parseInt(ceParam.substring(index1, index2));

		index1 = index2+1;
		index2 = ceParam.indexOf(',', index1);		
		int inPr = Integer.parseInt(ceParam.substring(index1, index2));

		index1 = index2+1;
		index2 = ceParam.indexOf(',', index1);		
		String qTypeStr = ceParam.substring(index1, index2);
		int mQWT = gis.getMaxQWT(qTypeStr);

		String paStr = ceParam.substring(index2+1, ceParam.length());
		PriceAuthority pa = gis.getPA(paStr);

		ComputingElement ce = new ComputingElement(id, secPerTU, pa,
							   qTypeStr,
							   cpuP, mQWT, cQWT,
							   inPr,
							   verbose, debug);

		if (verbose) {
		    System.out.print("registering CE"+id+": cpuPower="
				     +cpuP+", maxQWT="+mQWT
				     +", currQWT="+cQWT
				     +", initialPrice=");
		    if (inPr < 0)
			System.out.println("compute");
		    else
			System.out.println(""+inPr);

		    System.out.println("  queue type is '"+qTypeStr
				       +"', PA is '"+paStr+"'.");
		}

		gis.registerCE(ce);
	    } 
	    catch (Exception e) {
		System.err.println("CE"+id+" is not correctly defined in the "
				   + "configuration file!");
		System.exit(-1);
	    }
	}
    }


    // *********************
    // create job prototypes
    // *********************

    public void createJobPrototypes() {

	Integer numberJobTypes = conf.getIntegerValue("NUMBER_JOBTYPE");
	if (numberJobTypes == null) {
	    System.err.println("Number of job types not specified in conf "
			       + "file!");
	    System.exit(-1);
	}

	int nJTypes = numberJobTypes.intValue();
	for (int type = 0; type < nJTypes; type++) {
	    try {
		String jtParam = conf.getStringValue("JOBTYPE"+type);

		int index1 = 0, index2 = 0;
		index2 = jtParam.indexOf(',');
		int minCompEn = Integer.parseInt(jtParam.substring(index1,
								   index2));

		index1 = index2+1;
		index2 = jtParam.indexOf(',', index1);		
		int maxCompEn = Integer.parseInt(jtParam.substring(index1,
								   index2));

		index1 = index2+1;
		index2 = jtParam.indexOf(',', index1);		
		int minCPUReq = Integer.parseInt(jtParam.substring(index1,
								   index2));

		index1 = index2+1;
		index2 = jtParam.indexOf(',', index1);
		int maxCPUReq = Integer.parseInt(jtParam.substring(index1,
								   index2));

		String queueTypesStr = jtParam.substring(index2+1,
							 jtParam.length());

		// get ID strings of all accepted queue types:
		ArrayList acceptedQTypes = new ArrayList();

		if (queueTypesStr.equals("*")) { // all queue types accepted
		    acceptedQTypes = null;
		}
		else {
		    // get sinlge strings ...
		    index1 = 0;
		    index2 = queueTypesStr.indexOf('/');

		    String queueToken;
		    while(index2 != -1) { // still ID strings left
			queueToken = queueTypesStr.substring(index1, index2);

			if (gis.queueIsRegistered(queueToken))
			    acceptedQTypes.add(queueToken);
			else
			    System.err.println("Warning: JOBTYPE " + type
					       + ": ignoring unknown queue "
					       + "type '" + queueToken + "'.");

			index1 = index2+1;
			index2 = queueTypesStr.indexOf('/', index1);
		    }
		    // no more delimiting slashes found, this is the last queue
		    queueToken = queueTypesStr.substring(index1,
						     queueTypesStr.length());
		    if (gis.queueIsRegistered(queueToken))
			acceptedQTypes.add(queueToken);
		    else
			System.err.println("Warning: JOBTYPE " + type
					   + ": ignoring unknown queue "
					   + "type '" + queueToken + "'.");
		}

		Job jobPrototype = new Job(type, minCompEn, maxCompEn,
					   minCPUReq, maxCPUReq,
					   acceptedQTypes);

		gis.registerJobType(jobPrototype);

		if (verbose) {
		    System.out.println("registering job type " + type
				       + ": minCompEn=" + minCompEn
				       + ", maxCompEn=" + maxCompEn
				       + ", minCPUReq=" + minCPUReq
				       + ", maxCPUReq=" + maxCPUReq);
		    System.out.println("    accepted queue types = "
				       +acceptedQTypes);
		}

	    } 
	    catch (Exception e) {
		System.err.println("JOBTYPE"+type+" is not correctly defined "
				   + "in the configuration file!");
		System.exit(-1);
	    }
	}

    }


    // *************************
    // create core of simulation
    // *************************

    private void createSimCore() {

	String simCore = conf.getStringValue("SIMCORE");
	if (simCore == null) {
	    System.err.println("SIMCORE is not defined "
			       + "in the configuration file!");
	    System.exit(-1);
	}

	if (simCore.equals("AutoJobSubmSimCore")) {

	    // read conf file parameters for AutoJobSubmSimCore:    
	    int msBetwStdout, msBetwJobs;
	    int jobsPerSubm;
	    long msMaxRunTime;
	    String priceOutfilePrefix, qwtOutFilePrefix, jobsubOutFile;

	    Integer msWait = conf.getIntegerValue("MSEC_BETWEEN_STDOUT");
	    Integer msJobs = conf.getIntegerValue("MSEC_BETWEEN_JOBS");
	    Integer jSubm  = conf.getIntegerValue("JOBS_PER_SUBMISSION");
	    Double maxRTime = conf.getDoubleValue("MAX_RUNTIME");

	    priceOutfilePrefix = conf.getStringValue("PRICE_OUTFILE_PREFIX");
	    if (priceOutfilePrefix == null)
		priceOutfilePrefix = new String(DEF_PRICE_OUTFILE_PREFIX);

	    qwtOutFilePrefix = conf.getStringValue("QWT_OUTFILE_PREFIX");
	    if (qwtOutFilePrefix == null)
		qwtOutFilePrefix = new String(DEF_QWT_OUTFILE_PREFIX);

	    jobsubOutFile = conf.getStringValue("JOBSUBMISSION_OUTFILE");
	    if (jobsubOutFile == null)
		jobsubOutFile = new String(DEF_JOBSUBMISSION_OUTFILE);

	    if (msWait == null) {
		msBetwStdout = DEF_MSEC_BETWEEN_STDOUT;
		conf.addParameter("MSEC_BETWEEN_STDOUT",
				  new Integer(DEF_MSEC_BETWEEN_STDOUT));
	    }
	    else
		msBetwStdout = msWait.intValue();

	    if (msJobs == null) {
		msBetwJobs = DEF_MSEC_BETWEEN_JOBS;
		conf.addParameter("MSEC_BETWEEN_JOBS",
				  new Integer(DEF_MSEC_BETWEEN_JOBS));
	    }
	    else
		msBetwJobs = msJobs.intValue();

	    if ( (jSubm == null) || (jSubm.intValue()<1) ) {
		jobsPerSubm = DEF_JOBS_PER_SUBMISSION;
		conf.addParameter("JOBS_PER_SUBMISSION",
				  new Integer(DEF_JOBS_PER_SUBMISSION));
	    }
	    else
		jobsPerSubm = jSubm.intValue();

	    if (maxRTime == null || maxRTime.doubleValue()<=0.000001 ) {
		msMaxRunTime = 0; // unlimited
		conf.addParameter("MAX_RUNTIME", new Integer(0));
	    }
	    else
		// convert minutes to miliseconds
		msMaxRunTime = (long)(maxRTime.doubleValue() * 60000);


	    simulator = new AutoJobSubmSimCore(gis, rb, msMaxRunTime,
					       priceOutfilePrefix,
					       qwtOutFilePrefix, jobsubOutFile,
					       msBetwStdout, msBetwJobs,
					       jobsPerSubm,
					       verbose, debug);

	}
	else { // none of the known simulation cores
	    System.err.println("SIMCORE '"+simCore+"' is unknown! "
			       + "Check configuration file!");
	    System.exit(-1);
	}
    }

    // ****
    // main
    // ****

    public static void main(String[] args) {

	String confFName = null;
	if (args.length >= 1) {
	    if (args[0].equals("--help") || args[0].equals("-h"))
		helpMessage();
	    else
		confFName = args[0];
	}

	DGASSim sim = new DGASSim(confFName);

	// initiate simulation!
	sim.simulator.initSimulation();

	// run simulation!
	sim.simulator.runSimulation();

    }

}
