package org.edg.wl.dgas.dgassim.fabric;

import java.util.*;
import java.io.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.workload.pa.*;

public class ComputingElement {

    private boolean verbose;
    private boolean debug;

    private PriceAuthority pa;

    private int ceID;

    private String qTypeIdString;
    private int maxQWT; // maximum queue wait time in time units
    private int currQWT; // current queue wait time in time units

    private int cpuPower; // CPU power in UCE/time unit.

    private static double SEC_PER_TIME_UNIT;

    private long timeOfLastComp; // timestamp of last computation

    private int profit;

    // for data file creation:
    private FileOutputStream qwtOutFile, priceOutFile, profitOutFile;

    public ComputingElement(int id, double secPerTU, PriceAuthority pa,
			    String qTIdStr,
			    int cpuP, int mQWT, int cQWT,
			    int initialPrice,
			    boolean verbose, boolean debug) {

	SEC_PER_TIME_UNIT = secPerTU;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class ComputingElement");

	if (debug)
	    System.out.println("SEC_PER_TIME_UNIT = "+SEC_PER_TIME_UNIT);

	qTypeIdString = new String(qTIdStr);

	this.pa = pa;
	ceID = id;
	timeOfLastComp = (new Date()).getTime(); // current time
	maxQWT = mQWT;
	currQWT = cQWT;
	cpuPower = cpuP;

	profit = 0;

	// register CE to PA:
	if (pa != null) {
	    if (initialPrice < 0) // compute initial price
		pa.registerCE(this);
	    else // set initial price
		pa.registerCE(this, initialPrice);
	}
    } // constructor

    public FileOutputStream getQWTOutFile() {
	return qwtOutFile;
    }
    public FileOutputStream getPriceOutFile() {
	return priceOutFile;
    }
    public FileOutputStream getProfitOutFile() {
	return profitOutFile;
    }

    public void setQWTOutFile(FileOutputStream qwtOF) {
	qwtOutFile = qwtOF;
    }
    public void setPriceOutFile(FileOutputStream pOF) {
	priceOutFile = pOF;
    }
    public void setProfitOutFile(FileOutputStream pOF) {
	profitOutFile = pOF;
    }

    public void computeJobsInQueue() {
	long currTimestamp = (new Date()).getTime();

	synchronized (this) {
	    int seconds = ( (int)(currTimestamp - timeOfLastComp) / 1000 );

	    if ((double)seconds >= SEC_PER_TIME_UNIT) {
		int timeUnits = (int)(seconds / SEC_PER_TIME_UNIT);
		// the CPU power was already considered when queueing the job!

		currQWT -= timeUnits; // "compute" the jobs in queue.

		if (currQWT < 0)
		    currQWT = 0;

		timeOfLastComp += (int) (timeUnits * SEC_PER_TIME_UNIT * 1000);
	    }
	}
    }

    public boolean jobMayBeQueued(Job job){

	computeJobsInQueue();

	int timeUnits = (int) ((double)job.compEnergy()/(double)cpuPower +0.5);
	// how long does the job take to compute?

	if ( (maxQWT - currQWT) >= timeUnits ) { // queue not yet full
	    return true;
	}

	return false; // job too long for queue.
    }

    synchronized public boolean queueJob(Job job){

	computeJobsInQueue();

	int timeUnits = (int) ( (double)job.compEnergy() / (double)cpuPower
				+ 0.5);              // how long does the job
	                                             // take to compute?
	if ( (maxQWT - currQWT) >= timeUnits ) { // queue not yet full
	    currQWT += timeUnits;

	    // increase profit (take the price from the job info instead
	    // of the PA, since it might have changed (threads!). Jobs pay
	    // the price that was valid, when the CE was selcted.
	    // IMPORTANT: the profit made due to this job is the profit
	    // for the WHOLE execution time of the job!
	    profit += job.getJobPrice() * job.compEnergy();

	    job.setExecutingCE(ceID);

	    if (verbose)
		System.out.println("Queueing job " + job.dgJobId()+" of type "
				   + job.jobType() + ": " 
				   + job.compEnergy() +" UCE, req. CPU power: "
				   + job.minCPUPower() + "\n -> to CE" + ceID
				   + " with CPU power: " + cpuPower
				   + " UCE/time units.\n"
				   + " -> " + timeUnits + " time units added"
				   + " to CE's queue.");

	    return true;
	}

	return false; // not registered, queue too full.
    }

    public PriceAuthority getPA() {
	return pa;
    }

    public int getPrice() {
	return pa.getPrice(ceID);
    }

    public int getProfit() {
	return profit;
    }
    public void setProfit(int newProfit) {
	profit = newProfit;
    }

    public String getQTypeIdString() {
	return new String(qTypeIdString);
    }

    synchronized public int getCurrQWT() {
	computeJobsInQueue();
	return currQWT;
    }
    synchronized public int getCurrQWTNoComp() {
	return currQWT;
    }

    synchronized public int getMaxQWT() {
	return maxQWT;
    }

    synchronized public int getCPUPower() {
	return cpuPower;
    }

    synchronized public int getId() {
	return ceID;
    }

}
