package org.edg.wl.dgas.dgassim.workload.pa;

import java.util.*;
import java.io.*;

import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

public class DFPricingModule implements PricingModule {

    private boolean verbose;
    private boolean debug;

    TreeMap ceProfitMap;

    private double MIN_STEP_SIZE, MAX_STEP_SIZE;
    private double INIT_STEP_SIZE;
    private double STEP_INCREMENT, STEP_DECREMENT;
    private double DBLTURN_INCREMENT;

    private int MIN_PRICE;

    private static String PROFIT_OUTFILE_PREFIX;

    // constructors
    public DFPricingModule(int minPrice, double minStep, double maxStep,
			   double initStep, double stepIncr, double stepDecr,
			   double stepIncr2T, String profitOutfilePrefix,
			   boolean verbose, boolean debug) {

	MIN_PRICE = minPrice;
	MIN_STEP_SIZE = minStep;
	MAX_STEP_SIZE = maxStep;
	INIT_STEP_SIZE = initStep;
	STEP_INCREMENT = stepIncr;
	STEP_DECREMENT = stepDecr;
	DBLTURN_INCREMENT = stepIncr2T;

	PROFIT_OUTFILE_PREFIX = profitOutfilePrefix;

	this.verbose = verbose;
	this.debug = debug;

	ceProfitMap = new TreeMap();

	if (debug)
	    System.out.println("constructor of class DFPricingModule");

	if (debug)
	    System.out.println("MIN_PRICE=" + MIN_PRICE
			       + ", MIN_STEP_SIZE=" + MIN_STEP_SIZE
			       + ", MAX_STEP_SIZE=" + MAX_STEP_SIZE + "\n"
			       + "STEP_INCREMENT=" + STEP_INCREMENT
			       + ", STEP_DECREMENT=" + STEP_DECREMENT
			       + "\nDBLTURN_INCREMENT=" + DBLTURN_INCREMENT);
    }

    // internal class to contain profit information
    private class ProfitInfo {
	int profitPerSec;
	long lastTime; // last time of profit calculation, in miliseconds.
	long firstTime;

	// info for DF
	double stepSize;
	short direction;   // can be 1 or -1
	long noTurns;      // number of turns
	long noSteps;      // number of price steps
	long lastTurnStep; // step at which the last turn occured

	ProfitInfo(int pPS, long initTime) {
	    profitPerSec = pPS;
	    lastTime = initTime;
	    firstTime = initTime;

	    stepSize = INIT_STEP_SIZE;
	    direction = 1; // initially positive (increasing)
	    noTurns = 0;
	    noSteps = 0;
	    long lastTurnStep = Long.MIN_VALUE;
	}

	public void stepIncrement() {
	    stepSize *= STEP_INCREMENT;

	    if (stepSize > MAX_STEP_SIZE)
		stepSize = MAX_STEP_SIZE;
	}

	public void dblTurnIncrement() {
	    stepSize *= DBLTURN_INCREMENT;

	    if (stepSize > MAX_STEP_SIZE)
		stepSize = MAX_STEP_SIZE;
	}

	public void stepDecrement() {
	    stepSize /= STEP_DECREMENT;

	    if (stepSize < MIN_STEP_SIZE)
		stepSize = MIN_STEP_SIZE;
	}

    } // class ProfitInfo

    public int computePrice(ComputingElement ce) {

	int profitDifference = compareProfit(ce);

	int oldPrice = ce.getPrice();


	Integer ceID = new Integer(ce.getId());
	ProfitInfo pInfo = (ProfitInfo)ceProfitMap.get(ceID);

	// this is the next step:
	pInfo.noSteps++;

	System.out.println("CE="+ceID.intValue()+", old price="+oldPrice);
	System.out.println("CE="+ceID.intValue()+", step="+pInfo.noSteps
			   +", profitDifference="+profitDifference);
	System.out.print("CE="+ceID.intValue()+", old stepsize="
			   +pInfo.stepSize);

	if (profitDifference >= 0) { // profit doesn't decrease

	    if (pInfo.noTurns == 0) { // initial phase, always increment
		pInfo.stepIncrement();
		System.out.println(", new="+pInfo.stepSize
				   +" (initial phase, increment)");
	    }
	    else if ( pInfo.noTurns >= 1
		      && (pInfo.noSteps - pInfo.lastTurnStep > 5) ) {
		pInfo.stepIncrement(); // increment only five steps after turn
		System.out.println(", new="+pInfo.stepSize
				   +" (distant from turn, incr.)");
	    }
	    else {
		// otherwise keep step size fixed.
		System.out.println(", new="+pInfo.stepSize
				   +" (close to turn, no incr.)");
	    }

	    if (profitDifference == 0 && pInfo.profitPerSec == 0) {
		// if profit was zero at the last two steps, make sure we
		// decrease the price!!!
		if (pInfo.direction > 0) {
		    pInfo.direction *= -1;
		    pInfo.noTurns++;
		}
	    }
	}
	else { // profit decreases
	    pInfo.noTurns++;
	    pInfo.direction *= -1; // change direction of price adjustment

	    if (pInfo.noSteps == 2) {
		System.out.println(", new="+pInfo.stepSize
				   +" (second step, no decr.)");
		// don't change step size
	    }
	    else if ( pInfo.noTurns >= 2 
		      && (pInfo.noSteps == pInfo.lastTurnStep + 1) ) {
		pInfo.dblTurnIncrement(); // after two successive turns
		System.out.println(", new="+pInfo.stepSize
				   +" (double turn decr.)");
	    }
	    else {
		pInfo.stepDecrement();
		System.out.println(", new="+pInfo.stepSize
				   +" (decrement)");
	    }

	    pInfo.lastTurnStep = pInfo.noSteps;
	}

	int newPrice = oldPrice + (int) ((double)pInfo.direction
							* (pInfo.stepSize+0.5));
	if (newPrice < MIN_PRICE)
	    newPrice = MIN_PRICE;

	System.out.println("CE="+ceID.intValue()+", new price="+newPrice);

	return newPrice;
    }

    private int compareProfit(ComputingElement ce) {
	long currTime = (new Date()).getTime(); // in miliseconds
	Integer ceID = new Integer(ce.getId());

	// check whether we already have an entry for this CE!
	if (!ceProfitMap.containsKey(ceID)) {
	    synchronized (this) {
		// new entry
		ceProfitMap.put(ceID, new ProfitInfo(0, currTime));
		// initialize output file:
		try {
		    ce.setProfitOutFile(new FileOutputStream(
						    PROFIT_OUTFILE_PREFIX
						    +"_CE"+ce.getId()+".dat"));
		    ce.getProfitOutFile().write(
		       (new String("# Profit per second(GridCredits) of CE"
				   + ce.getId() + " (as function of time(s))\n"
				   + "# time Profit\n")).getBytes()
		       );
		} catch (Exception e) {
		    System.err.println("Cannot create profit output file for "
				       + "CE" + ce.getId());
		}
	    }

	}

	ProfitInfo pInfo;

	synchronized (this) {
	    pInfo = (ProfitInfo)ceProfitMap.get(ceID);
	}

	double seconds = ((double)currTime - (double)pInfo.lastTime)
	                                                     / (double)1000;

	int profitPerSec;
	if (seconds == 0)
	    profitPerSec = 0;
	else {
	    synchronized (this) {
		profitPerSec = (int) ((double)ce.getProfit() / seconds);
		ce.setProfit(0); // make sure that we begin to account anew.
	    }
	}

	int lastProfitPerSec = pInfo.profitPerSec;

	// update profit info:
	pInfo.profitPerSec = profitPerSec; // current profit
	pInfo.lastTime = currTime;         // current timestamp

	synchronized (this) {
	    // write new profit to output file
	    try {
		ce.getProfitOutFile().write(
			new String("" + ((currTime-pInfo.firstTime)/1000)
				   + " " + profitPerSec + "\n").getBytes()
			               );
	    } catch (Exception e) {
		System.err.println("Cannot write to profit output file for CE"
				   + ce.getId());
	    }
	}

	if (verbose) System.out.println("in compareProfit(ceId=" +ce.getId()
					+ "): seconds=" + seconds
					+ ", profitPerSec=" + profitPerSec
					+ ", last=" + lastProfitPerSec);

	System.out.println("CE=" + ceID.intValue() + ", old profit="
			   + lastProfitPerSec
			   + ", new profit=" + profitPerSec);

	return profitPerSec - lastProfitPerSec; // profit difference

    } // compareProfit() ...

}
