package org.edg.wl.dgas.dgassim.workload.pa;

import java.util.*;
import java.io.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

public class PriceAuthority extends Thread {

    private boolean verbose, debug;

    private int paID;
    private String paIDString;

    private ArrayList ceIDs;
    private TreeMap prices; // maps CE IDs to prices.
    private TreeMap ces; // maps CE IDs to CE references.

    private boolean runPA;

    private PricingModule pmod;

    private int PA_WAIT_TIME;
    private int PRICE_ADJ_THRES;

    private long timeOfLastComp; // timestamp of last price computation
    private long timeOfFirstComp; // timestamp of first price computation

    public PriceAuthority(int id, String paIdStr,
			  PricingModule pm, int paWaitTime,
			  int priceAdjThres,
			  boolean verbose, boolean debug) {
	super();

	paID = id;
	paIDString = new String(paIdStr);

	PA_WAIT_TIME = paWaitTime;
	PRICE_ADJ_THRES = priceAdjThres;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class PriceAuthority:\n"
			       + "PA_WAIT_TIME=" + PA_WAIT_TIME
			       + ", PRICE_ADJ_THRES=" + PRICE_ADJ_THRES);

	prices = new TreeMap();
	ces = new TreeMap();
	ceIDs = new ArrayList();
	pmod = pm;
	timeOfLastComp = (new Date()).getTime(); // current time
	timeOfFirstComp = timeOfLastComp;
    }

    public int getId() {
	return paID;
    }
    public String getIdString() {
	return new String(paIDString);
    }

    synchronized public int getPrice(int ceId) {
	return ((Integer)prices.get(new Integer(ceId))).intValue();
    }

    synchronized public void registerCE (ComputingElement ce, int price) {
	prices.put(new Integer(ce.getId()), new Integer(price));
	ces.put(new Integer(ce.getId()), ce);
	ceIDs.add(new Integer(ce.getId()));
    }
    // if price not specified
    synchronized public void registerCE (ComputingElement ce) {
	registerCE(ce, pmod.computePrice(ce));
    }

    public void stopPA() {
	runPA = false;
    }

    public void start() {
	runPA = true;
	if (verbose) System.out.println("Starting PA thread: " + paIDString);
	super.start();
    }

    public void run() {
	long secSinceFirst;

	FileOutputStream currFile;
	String currLine;

	ComputingElement currCE;
	Integer currCEId;

	while(runPA) {
	    // compute new prices if waitTime is over ...
	    long currTimestamp = (new Date()).getTime();

	    if (currTimestamp - timeOfLastComp >= PA_WAIT_TIME) {

		secSinceFirst = (currTimestamp-timeOfFirstComp)/1000;

		if (verbose)
		    System.out.println("[PA" + paID + " (" + paIDString + ")] "
				       + "sec=" + secSinceFirst + ", last="
				       + timeOfLastComp + ", now="
				       + currTimestamp);

		int size = ceIDs.size();
		int newPrice;
		int oldPrice;
		for (int i = 0; i < size; i++) {

		    currCEId = (Integer)ceIDs.get(i);
		    currCE = (ComputingElement)ces.get(currCEId);
		    oldPrice = getPrice(currCEId.intValue());
		    newPrice = pmod.computePrice(currCE);

		    if (verbose)
			System.out.println("[PA" + paID + " (" + paIDString
					   + ")] CE" + currCEId
					   + ": oldPrice=" + oldPrice
					   + ", newPrice=" + newPrice
					   + ", threshold=" + PRICE_ADJ_THRES);

		    // verify if price adjustment threshold is exceeded
		    if ( (newPrice - oldPrice) >= PRICE_ADJ_THRES ||
			 (newPrice - oldPrice) <= -PRICE_ADJ_THRES ) {

			if (verbose) System.out.println(" -> ... price "
							+ "set.");

			synchronized(this) {
			    // write price to file:
			    try {
				// all the price information
				currFile = currCE.getPriceOutFile();
				currLine = "" + secSinceFirst + " "
				           + newPrice+"\n";
				currFile.write(currLine.getBytes());
				currFile.flush();
				// all the qwt information
				currFile = currCE.getQWTOutFile();
				currLine = "" + secSinceFirst + " "
				           + currCE.getCurrQWTNoComp() +"\n";
				currFile.write(currLine.getBytes());
				currFile.flush();
			    } catch (Exception e) {
				System.err.println("\nCannot write data: "
						   + e.getMessage());
			    }

			    prices.put(currCEId, new Integer(newPrice));
			} // synchronized ...

		    }
		    else { // doesn't exceed threshold
			if (verbose) System.out.println(" -> ... price NOT"
							+ " set.");
		    }

		} // for

		timeOfLastComp = (new Date()).getTime();
	    } // if ...
	} // while

	return;
    }
}
